"""
This is a minimal working example of a backend application
implementing OAuth2 authorization through the "Authorization Code"
flow.

To run it, first set up the Python environment using Poetry or
or favourite Python environment tool.

Then run: python -m uvicorn main:app

Now to get an access token:

1. Open http://127.0.0.1:8000/authenticate in your browser
2. Create an account if you don't have any
3. Log in
4. Click "Authorize"
5. You should see "Success"

"""

import fastapi
from fastapi.responses import HTMLResponse
import random
import requests
import urllib.parse


client_id = 'YOUR_CLIENT_ID_HERE'
client_secret = 'YOUR_CLIENT_SECRET_HERE'
redirect_uri = 'http://127.0.0.1:8000/authorized/'  # for local development only

expected_state = None  # don't handle state like this in production, this is just for the example

app = fastapi.FastAPI()


# this endpoint represents the part of your system that prompts the end-user
# for granting access on his behalf
@app.get("/authenticate")
async def authenticate():
    global expected_state
    print(f'(1) authenticating')

    expected_state = random.randbytes(16).hex()  # some random state to prevent CSRF attacks

    # redirecting to ASI API for login
    return fastapi.responses.RedirectResponse(
        f"https://api.asi.swiss/oauth2/authorize/?"
        f"client_id={client_id}"
        f"&response_type=code"
        f"&redirect_uri={urllib.parse.quote(redirect_uri)}"
        f"&state={expected_state}",
        status_code=302
    )


# this endpoint is called by ASI API after the user successfully logs in and
# accepts the authorization
@app.get("/authorized")
async def callback(code: str, state: str):
    print(f'(2) app authorized: authcode={code}')

    assert state == expected_state

    print(f'(3) fetching access token on behalf of user')

    tokens = requests.post(
        'https://api.asi.swiss/oauth2/token/',
        data={
            'grant_type': 'authorization_code',
            'code': code,
            'redirect_uri': redirect_uri
        }, verify=False, allow_redirects=False,
        auth=(
            client_id, client_secret
        )
    ).json()

    # the result provides both an access token and a refresh token
    # use the refresh token from now on to renew the access token
    print(f'(3) access granted: token={tokens}')

    # now that we have an access token, we are free to access the
    # user's data and do anything with it

    # for the example here we will just show the user how many
    # devices he currently owns

    res = requests.get(
        f'https://api.asi.swiss/api/v1/devices/',
        params={'current': True},
        headers={'Authorization': f"Bearer {tokens['access_token']}"},
    ).json()

    n_devices = res['count']

    # this is an HTML example but you can do whatever you want after
    # receiving your access token
    html_content = f"""
    <html>
        <head>
            <title>Success</title>
        </head>
        <body>
            <h1>Success</h1>
            <p>You have successfully authorized this application to access your data on ASI API.</p>
            <p>You currently own {n_devices} devices.</p>
        </body>
    </html>
    """
    return HTMLResponse(content=html_content, status_code=200)
